;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;There are two fundamental modes of operation for Spawn,
;;sample-accurate and buffer-synchronized.
;;
;;Sample-accurate operation
;;
;;The problem is that we will probably be in the middle of a buffer when a
;;new event is spawned. The sub-network will still be operating at the same
;;buffer-frames, but offset by a constant value. So we need store for each
;;pad (1) the offset of the sub-network and (2) a pointer to the buffer
;;containing leftover data. Furthermore, as in general there will be more
;;than one SSSpawnAdder in a spawn element, there needs to be a
;;scheme-accessible way to make these pads and add them to the element.
;;
;;The solution will be to make a custom pad class, exposing the offset as a
;;property. The element will then be added to the element via
;;gst-element-add-pad. The spawn-adder will not be explicitly notified of
;;the new event, and won't listen to the "new-pad" signal; instead it must
;;loop through its pads on each iteration.
;;
;;Buffer-sunchronised operation
;;
;;This mode of operation is a bit easier, as well as more efficient. If
;;there is to be a new event within the buffer, it will start at the
;;beginning of the buffer. Thus the two networks will operate at the
;;same buffer-frames with zero relative offset.
;;
;;The spawner can be set to one mode or the other via an object property
;;on the spawn-adder.
;;
;;; Code:

(define-module (soundscrape spawn)
  :use-module (oop goops)
  :use-module (soundscrape element-support) ;; for support :)
  :use-module (soundscrape elements) ;; for the generics
  :use-module (soundscrape ugens) ;; for ss-spawn-adder
  :use-module (soundscrape gst-util)
  :use-module (soundscrape util)
  :use-module (scheme documentation)
  :use-module (gnome gstreamer) ;; <gst-bin>
  :export (spawn))

(use-modules ((srfi srfi-1) :select (circular-list zip)))

(define-class <spawn-class> ((class-of <gst-bin>)))

(define-class-with-docs <spawn> (<gst-bin>)
  ""
  (closure :param-spec `(,<gparam-boxed> :boxed-type ,<gclosure>))
  (proc :init-keyword :proc :init-value #f)
  (channels :param-spec `(,<gparam-int> :default-value 1 :minimum 1))
  (next-time :init-keyword :next-time :init-value 1)
  (max-repeats :init-keyword :max-repeats :init-value #f)
  (adders :init-value '() :accessor adders)
  (buffer-sync :init-value #f :init-keyword :buffer-sync)
  (timekeeper :init-value #f)
  :metaclass <spawn-class>)

(define-method (gobject:set-property (obj <spawn>) sym val)
  (next-method)
  (case sym
    ((channels)
     (or (not (slot-bound? obj 'adders)) (null? (adders obj)) (error "ak!" (adders obj)))
     (let ((pads (map (lambda args (audio ss-spawn-adder)) (make-list val))))
       (set! (adders obj) (map get-parent pads))
       (let ((timekeeper (get-parent (car pads))))
         (set timekeeper 'buffer-sync (slot-ref obj 'buffer-sync))
         (set timekeeper 'closure (slot-ref obj 'closure))
         (slot-set! obj 'timekeeper timekeeper))
       (for-each (lambda (o) (add obj o)) (adders obj))
       (for-each
        (lambda (pair)
          (add-ghost-pad obj (car pair) (format #f "src~A" (cadr pair))))
        (zip pads (iota val)))))))
     
(define spawn <spawn>)

(define-method (audio (class <spawn-class>) . args)
  (define (listify x) (if (list? x) x (list x)))
  (let-positional-keywords args
                           ((proc #f)
                            (channels 1)
                            (next-time 1)
                            (max-repeats #f))
    (or (and proc (procedure? proc))
        (error "Invalid :proc" proc))
    (or (and (integer? channels) (positive? channels))
        (error "Invalid :channels" channels))
    (or (or (not next-time) (procedure? next-time) (number? next-time))
        (error "Invalid :next-time" next-time))
    (or (or (not max-repeats) (and (integer? max-repeats) (positive? max-repeats)))
        (error "Invalid :max-repeats" max-repeats))
    (let ((instance (make class
                      :proc proc
                      :buffer-sync #t
                      :next-time next-time
                      :max-repeats max-repeats
                      :name #f)))
      ;; The closure is a thunk
      (set instance 'closure
           (make <gclosure>
             :func (let ((count 0))
                     (lambda ()
                       (let ((result (listify (proc instance count))))
                         (pk instance count)
                         (set! count (1+ count))
                         ;; reset the timer
                         (set (slot-ref instance 'timekeeper)
                              'time
                              (slot-ref instance 'next-time))
                         (if result
                             (begin
                               (let ((toplevels (graph-toplevel-elements result)))
                                 (for-each (lambda (e) (set-state e 'playing)) toplevels)
                                 (apply add instance toplevels))
                               (let lp ((pads result)
                                        (adders (apply circular-list (adders instance))))
                                 (if (not (null? pads))
                                     (begin
                                       (link (car pads) (get-pad (car adders) "sink%d"))
                                       (lp (cdr pads) (cdr adders))))))))))))
      (set instance 'channels channels) ;; actually triggers the work
      (get-pad-list instance))))
      

