/* Soundscrape
 * Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>
 *
 * ss-spawn-adder.c: A helper element for Spawn
 *
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License, or (at your option) any later version.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact:
 *
 * Free Software Foundation           Voice:  +1-617-542-5942
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
 * Boston, MA  02111-1307,  USA       gnu@gnu.org
 *
 * Some code based on gstadder.c from gst-plugins:
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *               2000 Wim Taymans <wim.taymans@chello.be>
 *               2001 Thomas <thomas@apestaart.org>
 */

#include <config.h>
#include "ss-spawn-adder.h"

static GstElementDetails spawn_adder_details = {
  "Spawn Adder",
  "Soundscrape/Controls",
  "A helper element for Spawn",
  "Andy Wingo <wingo at pobox dot com>",
};

enum {
    PROP_0,
    PROP_CLOSURE,
    PROP_TIME,
    PROP_BUFFER_SYNC
};

SS_PAD_TEMPLATE_SRC_FACTORY (ss_spawn_adder_src_factory, "src");

static GstPadTemplate*
ss_spawn_adder_sink_factory (gfloat default_value) 
{
    static GstPadTemplate *templ;
  
    if (!templ) {
        templ = ss_pad_template_sink_new ("sink%d", default_value);
        GST_PAD_TEMPLATE_PRESENCE (templ) = GST_PAD_REQUEST;
    }
  
    return templ;
}


static void 	ss_spawn_adder_class_init	(SSSpawnAdderClass *klass);
static void 	ss_spawn_adder_init		(SSSpawnAdder *this);
static void	ss_spawn_adder_set_property	(GObject *object, guint prop_id, 
                                                 const GValue *value, GParamSpec *pspec);
static void	ss_spawn_adder_get_property	(GObject *object, guint prop_id, 
                                                 GValue *value, GParamSpec *pspec);

static GstPad* 	ss_spawn_adder_request_new_pad 	(GstElement *element,
                                                 GstPadTemplate *temp,
                                                 const gchar *unused);

static void 	ss_spawn_adder_sync_loop	(GstElement *element);
static void 	ss_spawn_adder_accurate_loop	(GstElement *element);

static SSElementClass *parent_class = NULL;


SS_TYPE_MAKE_BOILERPLATE (ss_spawn_adder_get_type, SS_TYPE_ELEMENT,
                          "SSSpawnAdder", SSSpawnAdderClass, SSSpawnAdder,
                          ss_spawn_adder_class_init, ss_spawn_adder_init);

static void
ss_spawn_adder_class_init (SSSpawnAdderClass *klass) 
{
    GParamSpec *pspec;
    GObjectClass *oclass = (GObjectClass *) klass;
    GstElementClass *eclass = (GstElementClass *) klass;

    parent_class = g_type_class_ref (SS_TYPE_ELEMENT);

    oclass->set_property = ss_spawn_adder_set_property;
    oclass->get_property = ss_spawn_adder_get_property;

    pspec = g_param_spec_boxed ("closure", "closure", "A closure to create a new event",
                                G_TYPE_CLOSURE, G_PARAM_READWRITE);
    g_object_class_install_property (oclass, PROP_CLOSURE, pspec);

    pspec = g_param_spec_float ("time", "time", "The time until the next event. If negative,"
                                " the next event will come when the current one finishes.",
                                -G_MAXFLOAT, G_MAXFLOAT, 0.f, G_PARAM_READWRITE);
    g_object_class_install_property (oclass, PROP_TIME, pspec);

    pspec = g_param_spec_boolean ("buffer-sync", "buffer-sync", "Whether to synchronise new "
                                  "events with buffer boundaries. Otherwise events will be "
                                  "sample-accurate.",
                                  TRUE, G_PARAM_READWRITE | G_PARAM_CONSTRUCT);
    g_object_class_install_property (oclass, PROP_BUFFER_SYNC, pspec);

    eclass->request_new_pad = ss_spawn_adder_request_new_pad;

    ss_element_class_add_pad_template (eclass, ss_spawn_adder_src_factory ());
    ss_element_class_add_pad_template (eclass, ss_spawn_adder_sink_factory (0.0));
    gst_element_class_set_details (eclass, &spawn_adder_details);
}

static void 
ss_spawn_adder_init (SSSpawnAdder *this) 
{
    this->srcpad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_spawn_adder_src_factory ()));
    gst_element_add_pad (GST_ELEMENT (this), this->srcpad);
}

static void
ss_spawn_adder_set_property (GObject *object, guint prop_id,
                             const GValue *value, GParamSpec *pspec)
{
    SSSpawnAdder *this = SS_SPAWN_ADDER (object);

    switch (prop_id) {
    case PROP_CLOSURE:
        if (this->closure)
            g_closure_unref (this->closure);
        this->closure = g_value_dup_boxed (value);
        break;
    case PROP_TIME:
        this->time = g_value_get_float (value);
        break;
    case PROP_BUFFER_SYNC:
        this->buffer_sync = g_value_get_boolean (value);
        if (this->buffer_sync)
            gst_element_set_loop_function (GST_ELEMENT (object),
                                           ss_spawn_adder_sync_loop);
        else
            gst_element_set_loop_function (GST_ELEMENT (object),
                                           ss_spawn_adder_accurate_loop);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void 
ss_spawn_adder_get_property (GObject *object, guint prop_id,
                             GValue *value, GParamSpec *pspec)
{
    SSSpawnAdder *this = SS_SPAWN_ADDER (object);
   
    switch (prop_id) {
    case PROP_CLOSURE:
        g_value_set_boxed (value, this->closure);
        break;
    case PROP_TIME:
        g_value_set_float (value, this->time);
        break;
    case PROP_BUFFER_SYNC:
        g_value_set_boolean (value, this->buffer_sync);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static GstCaps *
ss_spawn_adder_pad_getcaps (GstPad *pad)
{
  SSSpawnAdder *this;
  const GstCaps *negotiated;
  
  this = SS_SPAWN_ADDER (GST_OBJECT_PARENT (pad));
  negotiated = gst_pad_get_negotiated_caps (this->srcpad);
  
  if (negotiated)
      return gst_caps_copy (negotiated);
  else
      return gst_caps_copy (gst_pad_get_pad_template_caps (this->srcpad));
}

static GstPad*
ss_spawn_adder_request_new_pad (GstElement *element, GstPadTemplate *templ, 
                                const gchar *unused) 
{
    static gint count = 0;
    SSSpawnAdder *this;
    GstPad *pad;
    gchar *name;

    this = SS_SPAWN_ADDER (element);

    pad = ss_pad_new_from_template (SS_PAD_TEMPLATE (ss_spawn_adder_sink_factory (0.0)));
    name = g_strdup_printf ("sink%d", count++);
    gst_object_set_name (GST_OBJECT (pad), name);
    g_free (name);
    
    gst_pad_set_getcaps_function (pad, ss_spawn_adder_pad_getcaps);

    gst_element_add_pad (GST_ELEMENT (this), pad);

    this->sinkpads = g_list_append (this->sinkpads, pad);
  
    return GST_PAD (pad);
}

/* if this->closure is set, then we are the keepers of the clock. our
   responsibility is to decrement this->time (which is in seconds) as processing
   goes on, and to invoke the closure when this->time<=zero. */

static void
ss_spawn_adder_sync_loop (GstElement *element)
{
    SSSpawnAdder *this = (SSSpawnAdder*)element;
    gint buffer_frames, i, frames_left, sample_rate;
    GstData *buffer_out = NULL;
    gfloat *data_out, *data_in;
    GList *l;

    SS_ELEMENT_ASSERT_NEGOTIATED (element);
    g_return_if_fail (SS_ELEMENT_IS_RATE (element, AUDIO));

    buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;
    sample_rate = SS_ELEMENT_CAST (this)->rate;

    /* see if we have any event first */
    if (this->closure) {
        frames_left = buffer_frames;
        while (this->time * sample_rate < frames_left) {
            frames_left -= this->time * sample_rate;
            this->time = 0;
            g_closure_invoke (this->closure, NULL, 0, NULL, NULL);
        }
        this->time -= frames_left * 1.f / sample_rate;
    }

    l = element->pads;
    while (l) {
        GstPad *pad = (GstPad*)l->data;
        GstData *buffer_in;

        if (pad == this->srcpad) {
            l = l->next;
            continue;
        }
        
        buffer_in = ss_pad_pull (SS_PAD_CAST (pad), buffer_frames);
        
        if (GST_IS_EVENT (buffer_in)) {
            l = l->next;
            if (GST_EVENT_TYPE (buffer_in) == GST_EVENT_EOS) {
                /* for eos, we keep going but disable the upstream network */
                gst_pad_unlink (pad, GST_PAD_PEER (pad));
                gst_element_remove_pad (element, pad);
            } else {
                gst_pad_push (this->srcpad, buffer_in);
            }
            continue;
        }
            
        if (!buffer_out) {
            if (SS_BUFFER_GET_FRAMES (buffer_in) == buffer_frames) {
                buffer_out = buffer_in;
                /* we keep the initial reference, and it already has the
                   data we want */
                l = l->next;
                continue;
            } else {
                buffer_out = ss_buffer_new_constant_val (0.0);
            }
        }
        i = MAX (buffer_frames, SS_BUFFER_GET_FRAMES (buffer_in));
        data_in = SS_BUFFER_DATA (buffer_in);
        data_out = SS_BUFFER_DATA (buffer_out);
        while (i--)
            *(data_out++) += *(data_in++);
        gst_data_unref (buffer_in);
        l = l->next;
    }
    if (!buffer_out)
        /* no sinks connected */
        buffer_out = ss_buffer_new_constant_val (0.0);
        
    ss_pad_push (SS_PAD_CAST (this->srcpad), buffer_out);
}

static void
ss_spawn_adder_accurate_loop (GstElement *element)
{
    SSSpawnAdder *this = (SSSpawnAdder*)element;
    gint buffer_frames, i;
    GstData *buffer_out = NULL;
    gfloat *data_out, *data_in;
    GList *l;

    SS_ELEMENT_ASSERT_NEGOTIATED (element);
    g_return_if_fail (SS_ELEMENT_IS_RATE (element, AUDIO));

    buffer_frames = SS_ELEMENT_CAST (this)->buffer_frames;

    /* see if we have any event first */

    /* we can process the existing sources now, then check the timer */

    for (l=this->sinkpads; l; l=l->next) {
        GstData *buffer_in = ss_pad_pull (SS_PAD_CAST (l->data), buffer_frames);
        if (!buffer_out) {
            if (SS_BUFFER_GET_FRAMES (buffer_in) == buffer_frames) {
                buffer_out = buffer_in;
                /* we keep the initial reference, and it already has the
                   data we want */
                continue;
            } else {
                buffer_out = ss_buffer_new_constant_val (0.0);
            }
        }
        i = MAX (buffer_frames, SS_BUFFER_GET_FRAMES (buffer_in));
        data_in = SS_BUFFER_DATA (buffer_in);
        data_out = SS_BUFFER_DATA (buffer_out);
        while (i--)
            *(data_out++) += *(data_in++);
        gst_data_unref (buffer_in);
    }
    if (!buffer_out)
        /* no sinks connected */
        buffer_out = ss_buffer_new_constant_val (0.0);
        
    ss_pad_push (SS_PAD_CAST (this->srcpad), buffer_out);
}

gboolean
init_ss_spawn_adder (GstPlugin *plugin)
{
    gst_element_register (plugin, "ss-spawn-adder", GST_RANK_NONE, SS_TYPE_SPAWN_ADDER);
    return TRUE;
}

